#!/usr/local/bin/python3

# S.D.G.

# Name: gglycanstring.py
# Description: gglycanstring is a simulator of the enzymes of ganglioside biosynthesis
# Author: Andrew G. McDonald
# ORCID iD: https://orcid.org/0000-0003-2727-176X
# Date: 7 November 2020

# Usage:
# python3 gglycanstring.py <iterations> [options]

# Options:
# 	-a	Use refined enzyme rules that match additional substrate-patterns
# 	-ln	Limit the number of GlcNAc residues to n per glycan (unused)
# 	-plist	Choose enzymes to knock out, as a comma-separated list of assignments of the form name=val, 
# 		where name is the short enzyme name and val is 0 (knockout) or 1 (knock-in).
# 		Example: -pST3Gal-V=0,beta4GalNAc-T1=0
# 	-sstring
#		Set the initial substrate to value of string (default is "T"). Example: -s"[S3]LT"


import sys, re

# define the enzymes used by the simulator as a tuple

enzymes = (
"UGCG",
"beta1Gal-T3",
"beta4Gal-T6",
"ST3Gal-V",
"ST8Sia-I",
"ST8Sia-V",
"beta4GalNAc-T1",
"beta3Gal-T4",
"ST3Gal-II",
"ST6GalNAc-V"
)


# set default parameter values

p = {}
for e in enzymes:
	p[e] = 1.0

limit = 0
# "advanced" mode activates further enzyme-substrate recognition patterns:
#  use command-line switch '-a' to enable
advanced = ""
halt = 0
# halt becomes true (1) if limit > 0 and c > limit
glycan = "T"

# parse the command line

if len(sys.argv)>1:
	maxiter = int(sys.argv[1])
	arglist = sys.argv[2:]
	for item in arglist:
		if item == '-a':
			advanced = '-a'
			print ("Using advanced model")
		if item.startswith('-l'):
			limit = int(item.replace('-l','',1))
			print ("Limiting to level", limit)
		if item.startswith('-p'):
			kolist = item.replace('-p','',1).split(",")
			for ep in kolist:
				pair = ep.split("=")
				p[pair[0]] = int(pair[1])
		if item.startswith('-s'):
			glycan = item.replace('-s','',1)

else:
	print ("No args supplied")
	maxiter = 1



# define the remaining hashes (dictionaries)

newp = {}
news = newp
pool = {}

newg = 0
newp[glycan] = glycan

if glycan not in pool:
	pool[glycan] = glycan
	newg += 1
else:
	newg = 0
n = {}
n[glycan] = len(pool)
product = ""
iteration = 0

# function definitions

def react( string, action ):
	global advanced
	if action == "UGCG":
		string = re.sub('^T', 'GT', string, 1)
	elif action == "beta1Gal-T3":
		string = re.sub('^T', 'LT', string, 1)
	elif action == "beta4Gal-T6":
		string = re.sub('^GT', 'L4GT', string, 1)
	elif action == "ST3Gal-V":
		string = re.sub('^L(?P<MATCH1>(4G)?)T', '[S3]L\g<MATCH1>T', string, 1)
	elif action == "ST8Sia-I":
		string = re.sub('^\[S3\]L4', '[S8S3]L4', string, 1)
	elif action == "ST8Sia-V":
		string = re.sub('^(?P<MATCH1>(\[)?)(?P<MATCH2>(S8)?)S3', '\g<MATCH1>S8\g<MATCH2>S3', string, 1)
	elif action == "beta4GalNAc-T1":
		string = re.sub('^(?P<MATCH1>(\[[S38]+\])?)L4', 'Vb4\g<MATCH1>L4', string, 1)
	elif action == "beta3Gal-T4":
		string = re.sub('^Vb4', 'L3Vb4', string, 1)
	elif action == "ST3Gal-II":
		string = re.sub('^L3Vb4', 'S3L3Vb4', string, 1)
	elif action == "ST6GalNAc-V":
		string = re.sub('S3L3Vb4', 'S3L3[S6]Vb4', string, 1)
	return string;

def chunk1( substrate, enzyme ):
	# for substrates with a single pattern-match
	global newg
	retstr = ""
	product = react(substrate,enzyme);
	if product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk1a( substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) with a single pattern-match
	global newg, newp, halt, limit
	retstr = ""
	product = react(substrate,enzyme);
	c = product.count('Y')
	if limit and limit > c :
		halt = 1
	if halt < 1 and product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		n[product] = len(pool)
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk2( regex, substrate, enzyme ):
	# for substrates having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l > 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				if nextproduct != '' and nextproduct != nextsubstrate :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1(substrate,enzyme)
	return

def chunk2a( regex, substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l > 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				c = product.count('Y')
				if limit and limit > c :
					halt = 1
				if halt < 1 and nextproduct != '' and nextsubstrate != nextproduct :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1a(substrate,enzyme)
	return

# start of main loop

while newg > 0 and iteration < maxiter :
	iteration += 1
	newg = 0

	news = newp	# the new products of one iteration become the substrates of the next
	newp = {}

	# cycle through the list of new substrates
	for substrate in news:

		# 1. Initiation: ceramide glucosyltransferase
		enzyme = "UGCG"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 2. Initiation: N-acylsphingosine galactosyltransferase
		enzyme = "beta1Gal-T3"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 3. Extension: glucosylceramide beta-1,4-galactosyltransferase
		enzyme = "beta4Gal-T6"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 4. Decoration with extension: lactosylceramide alpha-2,3-sialyltransferase
		enzyme = "ST3Gal-V"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 5. Extension: alpha-N-acetylneuraminate alpha-2,8-sialyltransferase
		enzyme = "ST8Sia-I"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 6. Termination: (alpha-2,8-N-acetylneuraminate alpha-2,8-sialyltransferase)
		enzyme = "ST8Sia-V"
		if p[enzyme] > 0:
			if not re.search('^\[S3\]L4', substrate):
				chunk1(substrate,enzyme);

		# 7. Extension: (N-acetylneuraminyl)-galactosylglucosylceramide N-acetylgalactosaminyltransferase
		enzyme = "beta4GalNAc-T1"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# 8. Extension: ganglioside galactosyltransferase
		enzyme = "beta3Gal-T4"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme)

		# 9. Termination with extension: (beta-1,3-galactosyl-ceramide alpha-2,3-sialyltransferase)
		enzyme = "ST3Gal-II"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme)

		# 10. Decoration: (alpha1,3-Sia-beta1,3-Gal-beta1,3-GalNAc alpha-2,6-sialyltransferase)
		enzyme = "ST6GalNAc-V"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme)

	if newg == 0 :
		print ("Nothing to be done!")
	else :
		print ("New structures: " + str(newg))

# print a list of all generated species

print ("\nGanglioside pool:")

for key in sorted(pool):
	print (key)

